from gates import new_var, background, FALSE, TRUE, AND, NOT, OR, XOR, IFF


def new_bv(width):
    return Bitvector(width, [new_var() for _ in range(width)])


def convert_int_to_bv(value):
    results = []
    while value > 0:
        results.append(TRUE() if value & 0b1  else FALSE())
        value = value >> 1
    return Bitvector(len(results), results[::-1])


class Bitvector:
    def __init__(self, width, body):
        self.content = body
        self.width = width

    # return the bit-vector's value
    def get_value(self, model):
        value = 0
        for i in range(self.width):
            var = self.content[i]
            if var in model:
                value += 1 << (self.width - 1 - i)

        return value


    def get_bit(self, i):
        return self.content[i]


def pad_bv(bv: Bitvector, pad_size: int):
    assert pad_size > 0
    return Bitvector(bv.width + pad_size, [FALSE() for _ in range(pad_size)] + bv.content)


def align_bvs(a: Bitvector, b: Bitvector):
    width_diff = a.width - b.width
    if width_diff > 0:
        b = pad_bv(b, width_diff)
    elif width_diff < 0:
        a = pad_bv(a, -width_diff)

    return a, b

def EQ(a: Bitvector, b: Bitvector, constraints = background):
    a, b = align_bvs(a, b)
    assert a.width == b.width
    width = a.width
    result = TRUE()
    for i in range(width):
        var_a = a.get_bit(i)
        var_b = b.get_bit(i)
        result = AND(result, IFF(var_a, var_b, constraints), constraints)
    return result

def NEQ(a, b, constraints = background):
    return NOT(EQ(a, b, constraints))


# define ripple carry-addition
def ADD(a: Bitvector, b: Bitvector, constraints = background):
    a, b = align_bvs(a, b)
    results = []
    width = a.width
    carry = FALSE()
    # starting from the least significant bit
    for i in range(width - 1, -1, -1):
        var_a = a.get_bit(i)
        var_b = b.get_bit(i)
        bit_output = XOR(carry, XOR(var_a, var_b, constraints), constraints)
        results.append(bit_output)
        carry = OR(OR(AND(carry, var_a, constraints), AND(carry, var_b, constraints), constraints),
                   AND(var_a, var_b, constraints), constraints)

    results.append(carry)
    return Bitvector(width + 1, results[::-1])



def GET(a: Bitvector, b: Bitvector, equal, constraints=background):
    a, b = align_bvs(a, b)
    assert a.width == b.width
    width = a.width
    ge = TRUE()
    gt = FALSE()
    for i in range(width):
        var_a = a.get_bit(i)
        var_b = b.get_bit(i)
        gt = OR(gt, AND(ge, AND(var_a, NOT(var_b), constraints), constraints), constraints)
        ge = AND(ge, OR(gt, NOT(AND(var_b, NOT(var_a), constraints)), constraints))

    return ge if equal else gt


def GE(a: Bitvector, b: Bitvector, constraints=background):
    return GET(a, b, True, constraints)


def GT(a: Bitvector, b: Bitvector, constraints=background):
    return GET(a, b, False, constraints)


def LT(a: Bitvector, b: Bitvector, constraints=background):
    return GT(b, a, constraints)


def LE(a: Bitvector, b: Bitvector, constraints=background):
    return GE(b, a, constraints)



