import sys

from pysat.formula import CNF

from bitvector import Bitvector, pad_bv, convert_int_to_bv, new_bv, EQ, ADD
from gates import TRUE, AND, IFF, OR, background
from sat_solver import solve


def _is_palindrome(bv: Bitvector, constraints =background):
    remaining_bits = bv.width
    results = TRUE()
    processed = 0
    while remaining_bits > 1:
        left = bv.get_bit(processed)
        right = bv.get_bit(-processed -1)
        # the bit from left and right are the same
        results = AND(results, IFF(left, right, constraints), constraints)
        remaining_bits -= 2
        processed += 1

    return results


def is_palindrome(bv: Bitvector, constraints  =background):
    result = _is_palindrome(bv, constraints)
    for i in range(1, bv.width):
        result = OR(result, _is_palindrome(pad_bv(bv, i), constraints), constraints)

    return result



def find_palindrome_sum(target: int):
    formula = background
    target_bv = convert_int_to_bv(target)
    bv1 = new_bv(target_bv.width)
    bv2 = new_bv(target_bv.width)
    formula.append([is_palindrome(bv1)])
    formula.append([is_palindrome(bv2)])
    sum = ADD(bv1, bv2)
    formula.append([EQ(sum, target_bv)])
    results = solve()
    if results:
        print("target {}".format(target_bv.get_value(results)))
        print("bv1 {} {}".format(bv1.get_value(results), bin(bv1.get_value(results))))
        print("bv2 {} {}".format(bv2.get_value(results),  bin(bv2.get_value(results))))

    # background.to_file("test")

if __name__ == "__main__":
    target = int(sys.argv[1])
    find_palindrome_sum(target)